/*
 *  CFRefHolder.h
 *	MLTERTFLib
 *
 *  Created by James Walker on 7/3/06.
 *
 *  Copyright (c) 2006 James W. Walker.
 *
 *	This software is provided 'as-is', without any express or implied warranty.
 *	In no event will the authors be held liable for any damages arising from the
 *	use of this software.
 *
 *	Permission is granted to anyone to use this software for any purpose,
 *	including commercial applications, and to alter it and redistribute it
 *	freely, subject to the following restrictions:
 *
 *	1. The origin of this software must not be misrepresented; you must not
 *	claim that you wrote the original software. If you use this software in a
 *	product, an acknowledgment in the product documentation would be appreciated
 *	but is not required.
 *
 *	2. Altered source versions must be plainly marked as such, and must not be
 *	misrepresented as being the original software.
 *
 *	3. This notice may not be removed or altered from any source distribution.
 */

#include <CoreFoundation/CoreFoundation.h>

/*!
	@class		CFRefHolder
	
	@abstract	Simple wrapper for a reference-counted CoreFoundation object.
	
	@discussion	Note that the constructor from a CF reference assumes that it
				is being given ownership of a reference, hence does not retain.
*/
template <class T>
class CFRefHolder
{
public:
						CFRefHolder()
							: mStr( NULL ) {}
						
	explicit			CFRefHolder( T* inStr )
							: mStr( inStr ) {}
						
						CFRefHolder( const CFRefHolder<T>& inOther )
							: mStr( inOther.mStr )
							{
								if (mStr != NULL)
								{
									CFRetain( mStr );
								}
							}
						
						~CFRefHolder()
							{
								if (mStr != NULL)
								{
									CFRelease( mStr );
								}
							}
	
	void				swap( CFRefHolder<T>& ioOther )
							{
								T*	tmp = mStr;
								mStr = ioOther.mStr;
								ioOther.mStr = tmp;
							}
						
	CFRefHolder&		operator=( const CFRefHolder<T>& inOther )
							{
								CFRefHolder<T>	temp( inOther );
								swap( temp );
								return *this;
							}
	
	T*					get() const
							{
								return mStr;
							}

private:
	T*					mStr;
};

typedef CFRefHolder<const struct __CFString>	CFStringRefHolder;
typedef CFRefHolder<struct __CFString>			CFMutableStringRefHolder;

typedef CFRefHolder<const struct __CFURL>		CFURLRefHolder;

