//
//  main.cpp
//  MLTE-RTF
//
//  Created by James Walker on 6/27/06.
//  Copyright ScriptPerfection Enterprises, Inc. 2006. All rights reserved.
//

#include <Carbon/Carbon.h>
#include "TApplication.h"
#include "TWindow.h"
#include "CreateRTFFromMLTE.h"
#include "LoadMLTEFromRTF.h"

#include <cstdio>

// Our custom application class
class CarbonApp : public TApplication
{
    public:
							
                            CarbonApp() {}
        virtual             ~CarbonApp() {}
        
    protected:
        virtual Boolean     HandleCommand( const HICommandExtended& inCommand );
	
};

// Our main window class
class MainWindow : public TWindow
{
    public:
                            MainWindow();
        virtual             ~MainWindow() {}
        
        static void         Create();
        
    protected:
        virtual Boolean     HandleCommand( const HICommandExtended& inCommand );

	private:
		void				SetSelStyle( Style inStyle );
		void				AppleCreateRTFFromMLTE();
		void				AppleLoadMLTEFromRTF();
		void				MyRTFFromMLTE();
		void				MyLoadMLTEFromRTF();
		void				ToggleSubSuperscript( bool inSuper );
		void				ToggleStrikethrough();
		
		HIViewRef			mStyledTextView;
		HIViewRef			mRTFTextView;
		
		TXNObject			mStyledTextTXN;
		TXNObject			mRTFTextTXN;
};

//--------------------------------------------------------------------------------------------
int main(int argc, char* argv[])
{
    CarbonApp app;
    
    // Create a new window. A full-fledged application would do this from an AppleEvent handler
    // for kAEOpenApplication.
    MainWindow::Create();
    
    app.Run();
    return 0;
}

//--------------------------------------------------------------------------------------------
Boolean
CarbonApp::HandleCommand( const HICommandExtended& inCommand )
{
    switch ( inCommand.commandID )
    {
        case kHICommandNew:
            MainWindow::Create();
            return true;
            
        // Add your own command-handling cases here
        
        default:
            return false;
    }
}

MainWindow::MainWindow()
	 : TWindow( CFSTR("MainWindow") )
	 , mStyledTextView( GetView( 'Text' ) )
	 , mRTFTextView( GetView( 'RTF ' ) )
	 , mStyledTextTXN( HITextViewGetTXNObject( mStyledTextView ) )
	 , mRTFTextTXN( HITextViewGetTXNObject( mRTFTextView ) )
{
	
}

//--------------------------------------------------------------------------------------------
void
MainWindow::Create()
{
    MainWindow* wind = new MainWindow();

    // Position new windows in a staggered arrangement on the main screen
    RepositionWindow( *wind, NULL, kWindowCascadeOnMainScreen );
    
    // The window was created hidden, so show it
    wind->Show();
}

void	MainWindow::SetSelStyle( Style inStyle )
{
	TXNTypeAttributes	attribute;
	
	attribute.tag	= kTXNQDFontStyleAttribute;
	attribute.size	= kTXNQDFontStyleAttributeSize;
	attribute.data.dataValue = inStyle;
	
	TXNSetTypeAttributes( mStyledTextTXN, 1, &attribute,
											  kTXNUseCurrentSelection,
											  kTXNUseCurrentSelection );
}

void	MainWindow::ToggleSubSuperscript( bool inSuper )
{
	TXNOffset	selStart, selEnd;
	TXNGetSelection( mStyledTextTXN, &selStart, &selEnd );
	
	ItemCount	runCount, run;
	TXNCountRunsInRange( mStyledTextTXN, selStart,
		selEnd, &runCount );
	
	TXNTypeAttributes	atts[2];
	atts[0].tag = kATSUCrossStreamShiftTag;
	atts[0].size = sizeof(Fixed);
	atts[1].tag = kTXNQDFontSizeAttribute;
	atts[1].size = kTXNFontSizeAttributeSize;
	
	for (run = 0; run < runCount; ++run)
	{
		TXNOffset	runStart, runEnd;
		TXNDataType	runType;
		
		OSStatus	err = TXNGetIndexedRunInfoFromRange( mStyledTextTXN, run,
			selStart, selEnd,
			&runStart, &runEnd, &runType,
			2, atts );
		
		if ( (err == noErr) and (runType == kTXNUnicodeTextData) )
		{
			if (runStart < selStart)
			{
				runStart = selStart;
			}
			if (runEnd > selEnd)
			{
				runEnd = selEnd;
			}
			Fixed oldFixedSize = static_cast<Fixed>( atts[1].data.dataValue );
			double oldDoubleSize = Fix2X( oldFixedSize );
			Fixed oldFixedShift = static_cast<Fixed>( atts[0].data.dataValue );
			double	newShift = oldDoubleSize * 0.25;
			if (not inSuper)
			{
				newShift = - newShift;
			}
			
			if (oldFixedShift == 0)
			{
				atts[0].data.dataValue = static_cast<UInt32>(
					X2Fix( newShift ) );
				atts[1].data.dataValue = static_cast<UInt32>(
					X2Fix( oldDoubleSize / 1.2 ) );
			}
			else if ((oldFixedShift > 0) == inSuper)	// already shifted
			{
				atts[0].data.dataValue = 0;
				atts[1].data.dataValue = static_cast<UInt32>(
					X2Fix( oldDoubleSize * 1.2 ) );
			}
			else	// was shifted the opposite way; leave font size alone
			{
				atts[0].data.dataValue = static_cast<UInt32>(
					X2Fix( newShift ) );
			}
			err = TXNSetTypeAttributes( mStyledTextTXN, 2, atts,
				runStart, runEnd );
			if (err != noErr)
			{
				std::printf( "Error %d\n", err );
			}
		}
	}
	HIViewSetNeedsDisplay( mStyledTextView, true );
}

/*
	From an ATSUStyle, I can get
	kATSUStyleStrikeThroughTag
	kATSUFontTag
	kATSUSizeTag
	kATSUColorTag
	kATSUCrossStreamShiftTag
	
	but I will not get QD style
*/

void	MainWindow::ToggleStrikethrough()
{
	TXNOffset	selStart, selEnd;
	TXNGetSelection( mStyledTextTXN, &selStart, &selEnd );
	
	ItemCount	runCount, run;
	TXNCountRunsInRange( mStyledTextTXN, kTXNUseCurrentSelection,
		kTXNUseCurrentSelection, &runCount );
	
	TXNTypeAttributes	atts;
	atts.tag = kTXNATSUIStyle;
	atts.size = kTXNATSUIStyleSize;
	atts.data.dataValue = 0;
	
	for (run = 0; run < runCount; ++run)
	{
		TXNOffset	runStart, runEnd;
		TXNDataType	runType;
		
		OSStatus	err = TXNGetIndexedRunInfoFromRange( mStyledTextTXN, run,
			kTXNUseCurrentSelection, kTXNUseCurrentSelection,
			&runStart, &runEnd, &runType,
			1, &atts );
		
		if ( (err == noErr) and (runType == kTXNUnicodeTextData) )
		{
			if (runStart < selStart)
			{
				runStart = selStart;
			}
			if (runEnd > selEnd)
			{
				runEnd = selEnd;
			}
			
			ATSUStyle	prevStyle = static_cast<ATSUStyle>(
				atts.data.dataPtr );
			ATSUStyle	workStyle = NULL;
			ATSUCreateAndCopyStyle( prevStyle, &workStyle );
			
			ByteCount	actualSize;
			Boolean		theFlag;
			err = ATSUGetAttribute( workStyle, kATSUStyleStrikeThroughTag,
				sizeof(Boolean), &theFlag, &actualSize );

			theFlag = not theFlag;
			ATSUAttributeTag	theTag = kATSUStyleStrikeThroughTag;
			ATSUAttributeValuePtr	theValue = &theFlag;
			err = ATSUSetAttributes( workStyle, 1, &theTag, &actualSize,
				&theValue );
			
			atts.data.dataValue = reinterpret_cast<UInt32>( workStyle );
			
			err = TXNSetTypeAttributes( mStyledTextTXN, 1, &atts,
				runStart, runEnd );
			if (err != noErr)
			{
				std::printf( "Error %d\n", err );
			}
			ATSUDisposeStyle( workStyle );
		}
	}
	HIViewSetNeedsDisplay( mStyledTextView, true );
}


//--------------------------------------------------------------------------------------------
Boolean
MainWindow::HandleCommand( const HICommandExtended& inCommand )
{
	Boolean	handled = false;
	
    switch ( inCommand.commandID )
    {
        // Add your own command-handling cases here
		case 'Bold':
			SetSelStyle( bold );
			handled = true;
			break;
		
		case 'Ital':
			SetSelStyle( italic );
			handled = true;
			break;
			
		case 'Undr':
			SetSelStyle( underline );
			handled = true;
			break;
		
		case 'Supr':
			ToggleSubSuperscript( true );
			break;
			
		case 'Subs':
			ToggleSubSuperscript( false );
			break;
			
		case 'Stri':
			ToggleStrikethrough();
			break;
			
		case 'CRTF':
			AppleCreateRTFFromMLTE();
			handled = true;
			break;
			
		case 'RRTF':
			AppleLoadMLTEFromRTF();
			handled = true;
			break;
		
		case '>RTF':
			MyRTFFromMLTE();
			break;
		
		case '<RTF':
			MyLoadMLTEFromRTF();
			break;
		
        default:
            return false;
    }
	
	return handled;
}

void	MainWindow::AppleCreateRTFFromMLTE()
{
	CFDataRef	dataCF = NULL;
	OSStatus	err = ::TXNFlattenObjectToCFDataRef( mStyledTextTXN,
		kTXNRichTextFormatData, &dataCF );
	if (err == noErr)
	{
		CFIndex	len = ::CFDataGetLength( dataCF );
		const UInt8*	dataPtr = ::CFDataGetBytePtr( dataCF );
		
		err = TXNSetData( mRTFTextTXN, kTXNTextData, dataPtr, len,
			kTXNStartOffset, kTXNEndOffset );
		
		CFRelease( dataCF );
	}
	HIViewSetNeedsDisplay( mRTFTextView, true );
}


static CFURLRef	CreateTempFileFromData( CFDataRef inData )
{
	FSRef	dirRef;
	FSFindFolder( kUserDomain, kChewableItemsFolderType, kCreateFolder, &dirRef );
	CFURLRef	dirURL = CFURLCreateFromFSRef( NULL, &dirRef );
	CFURLRef	fileURL = CFURLCreateCopyAppendingPathComponent( NULL, dirURL,
		CFSTR("MLTE-test.rtf"), false );
	CFRelease( dirURL );
	SInt32	errCode;
	CFURLWriteDataAndPropertiesToResource( fileURL, inData, NULL, &errCode );
	return fileURL;
}

void	MainWindow::AppleLoadMLTEFromRTF()
{
	CFDataRef	dataCF = NULL;
	OSStatus	err = ::TXNFlattenObjectToCFDataRef( mRTFTextTXN,
		kTXNTextData, &dataCF );
	if (err == noErr)
	{
		CFURLRef	fileURL = CreateTempFileFromData( dataCF );
		
		CFTypeRef	theKeys[] = {
			kTXNDataOptionDocumentTypeKey
		};
		CFTypeRef	theValues[] = {
			kTXNRTFDocumentType
		};
		CFDictionaryRef	optDict = CFDictionaryCreate( NULL, theKeys, theValues,
			1, &kCFTypeDictionaryKeyCallBacks, &kCFTypeDictionaryValueCallBacks );
		
		TXNReadFromCFURL( mStyledTextTXN, kTXNStartOffset, kTXNEndOffset,
			optDict, fileURL, NULL );
		
		CFRelease( optDict );
		CFRelease( fileURL );
		CFRelease( dataCF );
	}
	HIViewSetNeedsDisplay( mStyledTextView, true );
}

void	MainWindow::MyRTFFromMLTE()
{
	CFDataRef	dataCF = ::CreateRTFFromMLTE( mStyledTextTXN );
	if (dataCF != NULL)
	{
		CFIndex	len = ::CFDataGetLength( dataCF );
		const UInt8*	dataPtr = ::CFDataGetBytePtr( dataCF );
		
		TXNSetData( mRTFTextTXN, kTXNTextData, dataPtr, len,
			kTXNStartOffset, kTXNEndOffset );
		
		CFRelease( dataCF );
	}
	else
	{
		TXNSetData( mRTFTextTXN, kTXNTextData, NULL, 0,
			kTXNStartOffset, kTXNEndOffset );
	}
	HIViewSetNeedsDisplay( mRTFTextView, true );
}

void	MainWindow::MyLoadMLTEFromRTF()
{
	CFDataRef	dataCF = NULL;
	OSStatus	err = ::TXNFlattenObjectToCFDataRef( mRTFTextTXN,
		kTXNTextData, &dataCF );
	if (err == noErr)
	{
		::LoadMLTEFromRTF( dataCF, mStyledTextTXN );
		
		CFRelease( dataCF );
	}
	HIViewSetNeedsDisplay( mStyledTextView, true );
}


